/*
 *	bmp2fnt.c - Convert windows BMP to GBA font format
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include <xfont.h>

#define NBPB	8

#define ROUNDUP4(x)	(((x) + 3) & ~3)
#define ROUNDUPBYTE(x)	(((x) + 7) & ~7)

#ifdef DEBUG
#define DPRINTF	if (Verbose) printf
#else
#define printf
#endif

static char	*cmdname = "bmp2fnt";

static char	*Usage[] = {
	"Usage: %s [options] {bmp-file} {fnt-file}",
	"	-sW,H:	font width and height",
	"	-cS,E:	start,end code (in hex)",
	"	-oS,E:	start,end offset (in hex)",
	"	-pX:	proportional font and width for default char",
	"	-gX,Y:	x,y font gap",
	"	-nNNN:	font name (length must be 3 bytes)",
	"	-t#:	CCT index (#=0-255)",
	"	-v:	verbose output",
	NULL,
};

static int	FontWidth;
static int	FontHeight;
static int	StartCode, EndCode;
static int	StartOffset, EndOffset;
static int	FontFlags;
static int	DfltFontWidth;
static int	xGap, yGap;
static BYTE	FontName[XF_NAMELEN+1];
static int	CCTindex;
static BOOL	Verbose;

/*
 *	fatal
 */
static void fatal(char *s)
{
	fprintf(stderr, "%s: ", cmdname);
	perror(s);
	exit(1);
}

/*
 *	error
 */
static void error(char *fmt, void *v)
{
	fprintf(stderr, "%s: ", cmdname);
	fprintf(stderr, fmt, v);
	fprintf(stderr, "\n");
	exit(1);
}

/*
 *	BuildFontHeader
 */
static void BuildFontHeader(PXFONTHEAD pxf)
{
	memcpy(pxf->sig, XFONT_SIG, sizeof(pxf->sig));
	pxf->ver = XFONT_VER;
	pxf->size = sizeof(*pxf);
	strcpy(pxf->name, FontName);
	pxf->flags = FontFlags;
	pxf->unit = (FontFlags & XFF_PROP) ?
		ROUNDUPBYTE(FontWidth * FontHeight) / NBPB + sizeof(BYTE) :
		FontWidth * FontHeight;
	pxf->width = FontWidth;
	pxf->height = FontHeight;
	pxf->xgap = xGap;
	pxf->ygap = yGap;
	pxf->scode = StartCode;
	pxf->ecode = EndCode;
	pxf->cct = CCTindex;
	memset(pxf->reserved, 0, sizeof(pxf->reserved));
}

/*
 *	GetFontPos
 */
static void GetFontPos(PBYTE pb0, int bmpwidth, int *pxstart, int *pxend)
{
	PBYTE		pb;
	int		xstart, xend;
	int		x, y;

	/*
	 *	Find the start position
	 */

	for (x = 0; x < FontWidth; x++) {
		pb = &pb0[x];
		for (y = 0; y < FontHeight; y++) {
			if (*pb) {
				*pxstart = x;
				goto found;
			}
			pb -= bmpwidth;
		}
	}

	/*
	 *	All clear
	 */

	*pxstart = 0;
	*pxend = DfltFontWidth;
	return;

found:

	/*
	 *	Find the end position
	 */

	for (x = FontWidth - 1; x >= 0; x--) {
		pb = &pb0[x];
		for (y = 0; y < FontHeight; y++) {
			if (*pb) {
				*pxend = x + 1;
				return;
			}
			pb -= bmpwidth;
		}
	}

	error("%s: internal error.", __FILE__);
}

/*
 *	bmp2fnt
 */
static void bmp2fnt(char *bmpname, char *fntname)
{
	FILE			*fpbmp;
	FILE			*fpfnt;
	PBYTE			pbmp, pfnt;
	PBYTE			pb0, pb1, pb;
	PBYTE			pf, pf0;
	int			bmpsize, fntsize, palsize;
	int			ncols, nlines;
	int			nbmps, nfonts;
	int			bmpwidth, bmpheight, bmpnext;
	int			off;
	int			x, y;
	int			xstart, xend;
	int			w;
	BYTE			bmask;
	BITMAPFILEHEADER	bmfh;
	BITMAPINFOHEADER	bmih;

	/*
	 *	Check bitmap file and params
	 */

	if ((fpbmp = fopen(bmpname, "rb")) == NULL)
		fatal(bmpname);

	if (fread(&bmfh, sizeof bmfh, 1, fpbmp) != 1)
		fatal(bmpname);

	if (bmfh.bfType != 'MB')
		error("%s: Not a bitmap file.", bmpname);

	if (fread(&bmih, sizeof bmih, 1, fpbmp) != 1)
		fatal(bmpname);

	if (bmih.biSize != sizeof bmih)
		error("%s: Invalid BMIH.", bmpname);

	if (Verbose) {
		printf("biWidth = %d\n", bmih.biWidth);
		printf("biHeight = %d\n", bmih.biHeight);
		printf("biPlanes = %d\n", bmih.biPlanes);
		printf("biBitCount = %d\n", bmih.biBitCount);
	}

	if (bmih.biBitCount != 8)
		error("%s: must be 8bpp at this time.", bmpname);

	if (!FontWidth || bmih.biWidth % FontWidth)
		error("%s: must be divided with font width.", bmpname);

	if (!FontHeight || bmih.biHeight % FontHeight)
		error("%s: must be divided with font height.", bmpname);

	ncols = bmih.biWidth / FontWidth;
	nlines = bmih.biHeight / FontHeight;
	nbmps = ncols * nlines;
	DPRINTF("ncols = %d\n", ncols);
	DPRINTF("nlines = %d\n", nlines);
	DPRINTF("nbmps = %d\n", nbmps);

	if (StartOffset < 0 || StartOffset >= nbmps)
		error("start offset: out of range.", NULL);
	if (EndOffset < 0 || EndOffset >= nbmps)
		error("end offset: out of range.", NULL);
	if (StartOffset > EndOffset)
		error("start offset > end offset.", NULL);

	nfonts = (EndOffset - StartOffset) + 1;

	/*
	 *	Load bitmap
	 */

	palsize = sizeof(DWORD) * 256;
	bmpsize = bmfh.bfSize - sizeof(bmfh) - sizeof(bmih) - palsize;
	DPRINTF("bmpsize = %d\n", bmpsize);

	if ((pbmp = (PBYTE)malloc(bmpsize)) == NULL)
		fatal("");

	if (fseek(fpbmp, palsize, SEEK_CUR))	// skip palette table
		fatal(bmpname);

	if (fread(pbmp, bmpsize, 1, fpbmp) != 1)
		fatal(bmpname);

	if (FontFlags & XFF_PROP)
		fntsize = ((ROUNDUPBYTE(FontWidth * FontHeight) / NBPB) +
			sizeof(BYTE)) * nfonts + sizeof(XFONTHEAD);
	else
		fntsize = ROUNDUPBYTE(FontWidth * FontHeight * nfonts)
			/ NBPB + sizeof(XFONTHEAD);
	if ((pfnt = (PBYTE)malloc(fntsize)) == NULL)
		fatal("");
	memset(pfnt, 0, fntsize);
	DPRINTF("fntsize = %d\n", fntsize);

	/*
	 *	Build font image
	 */

	BuildFontHeader((PXFONTHEAD)pfnt);

	bmpwidth = ROUNDUP4(bmih.biWidth);
	bmpheight = bmih.biHeight;
	bmpnext = bmpwidth * FontHeight;
	DPRINTF("bmpwidth = %d\n", bmpwidth);
	DPRINTF("bmpheight = %d\n", bmpheight);
	DPRINTF("bmpnext = %d\n", bmpnext);

	pb0 = &pbmp[bmpsize - bmpwidth];
	pf = pfnt + sizeof(XFONTHEAD);
	if (!(FontFlags & XFF_PROP)) {
		bmask = 0x80;
		xstart = 0;
		xend = FontWidth;
	}

	for (off = StartOffset; off <= EndOffset; off++) {
#ifdef DEBUG
		int	temp;
		DPRINTF("off = %04x\n", off);
		temp = ((off / ncols) * bmpnext) +
			((off % ncols) * FontWidth);
		DPRINTF("temp = %d\n", temp);
		DPRINTF("font off = %06x , bit mask = %02x\n",
			pf - pfnt, bmask);
#endif
		pb1 = pb0 - ((off / ncols) * bmpnext) +
			((off % ncols) * FontWidth);

		if (FontFlags & XFF_PROP) {
			GetFontPos(pb1, bmpwidth, &xstart, &xend);
			w = xend - xstart;
			*pf++ = w;	// font width
			pf0 = pf;
			bmask = 0x80;
			DPRINTF("xstart,xend = %d,%d\n", xstart, xend);
			DPRINTF("w = %d\n", w);
		}

		for (y = 0; y < FontHeight; y++) {
			pb = pb1 + xstart;
			if (pb < pbmp || pb >= &pbmp[bmpsize])
				error("Invalid pb.", NULL);
			for (x = xstart; x < xend; x++) {
				if (pf < pfnt || pf >= &pfnt[fntsize])
					error("Invalid pf.", NULL);
				DPRINTF("%c", *pb ? 'o' : ' ');
				*pf |= (*pb++) ? bmask : 0;
				if ((bmask >>= 1) == 0) {
					bmask = 0x80;
					pf++;
				}
			}
			pb1 -= bmpwidth;
			DPRINTF("\n");
		}

		if (FontFlags & XFF_PROP)
			pf = pf0 + ROUNDUPBYTE(FontWidth * FontHeight) / NBPB;

		DPRINTF("\n");
	}

	/*
	 *	Build font file
	 */

	if ((fpfnt = fopen(fntname, "wb")) == NULL)
		fatal(fntname);
	if (fwrite(pfnt, fntsize, 1, fpfnt) != 1)
		fatal(fntname);
	fclose(fpfnt);

	/*
	 *	Finishing
	 */

	free(pfnt);
	free(pbmp);

	fclose(fpbmp);
}

/*
 *	usage
 */
static void usage()
{
	char		**p;

	p = Usage;
	fprintf(stderr, *p++, cmdname);
	fprintf(stderr, "\n");
	for (; *p; p++)
		fprintf(stderr, "%s\n", *p);

	exit(1);
}

/*
 *	CheckCodeRange
 */
static BOOL CheckCodeRange(int scode, int ecode)
{
	return (scode >= StartCode && scode <= EndCode &&
		ecode >= StartCode && ecode <= EndCode);
}

/*
 *	main
 */
void main(int argc, char* argv[])
{
	FontFlags = XFF_FIXED;

	/*
	 *	Get options
	 */

	argc--, argv++;
	while (argc && **argv == '-') switch (argv[0][1]) {
	default:
		usage();
		// NOTREACHED
	case 's':
		sscanf(&argv[0][2], "%d,%d", &FontWidth, &FontHeight);
		argc--, argv++;
		break;
	case 'c':
		sscanf(&argv[0][2], "%x,%x", &StartCode, &EndCode);
		argc--, argv++;
		break;
	case 'o':
		sscanf(&argv[0][2], "%x,%x", &StartOffset, &EndOffset);
		argc--, argv++;
		break;
	case 'p':
		FontFlags |= XFF_PROP;
		DfltFontWidth = atoi(&argv[0][2]);
		argc--, argv++;
		break;
	case 'g':
		sscanf(&argv[0][2], "%d,%d", &xGap, &yGap);
		argc--, argv++;
		break;
	case 'n':
		if (strlen(&argv[0][2]) != XF_NAMELEN)
			usage();
		strcpy(FontName, &argv[0][2]);
		argc--, argv++;
		break;
	case 't':
		if (argv[0][2] == '\0')
			usage();
		CCTindex = atoi(&argv[0][2]);
		if (CCTindex < 0 || CCTindex > 255)
			usage();
		argc--, argv++;
		break;
	case 'v':
		Verbose = TRUE;
		argc--, argv++;
		break;
	}

	if (Verbose) {
		printf("FontWidth = %d\n", FontWidth);
		printf("FontHeight = %d\n", FontHeight);
		printf("Code Range(%x,%x)\n", StartCode, EndCode);
		printf("Offset Range(%x,%x)\n", StartOffset, EndOffset);
		printf("\n");
	}

	if (argc != 2)
		usage();

	bmp2fnt(argv[0], argv[1]);
}

